#pragma once

#include "vector3.h"

#include <ostream>

template <typename T>
class Matrix3x3 {
    public:
        Matrix3x3() = default;
        Matrix3x3(T s) : m_rows{
            {s, static_cast<T>(0), static_cast<T>(0)},
            {static_cast<T>(0), s, static_cast<T>(0)},
            {static_cast<T>(0), static_cast<T>(0), s}
        } {}
        Matrix3x3(Vector3<T> r1, Vector3<T> r2, Vector3<T> r3) : m_rows{r1, r2, r3} {}
        Matrix3x3(const Matrix3x3<T> &other) = default;

        Matrix3x3<T> operator-() const { return Matrix3x3<T>(-m_rows[0], -m_rows[1], -m_rows[2]); }

        Matrix3x3<T> &operator=(const Matrix3x3<T> &other) {
            m_rows[0] = other.m_rows[0];
            m_rows[1] = other.m_rows[1];
            m_rows[2] = other.m_rows[2];
            return *this;
        }

        Matrix3x3<T> operator+(const Matrix3x3<T> &other) const {
            return Matrix3x3(
                m_rows[0] + other.m_rows[0],
                m_rows[1] + other.m_rows[1],
                m_rows[2] + other.m_rows[2]
            );
        }
        Matrix3x3<T> &operator+=(const Matrix3x3<T> &other) {
            m_rows[0] += other.m_rows[0];
            m_rows[1] += other.m_rows[1];
            m_rows[2] += other.m_rows[2];
            return *this;
        }

        Matrix3x3<T> operator-(const Matrix3x3<T> &other) const {
            return Matrix3x3(
                m_rows[0] - other.m_rows[0],
                m_rows[1] - other.m_rows[1],
                m_rows[2] - other.m_rows[2]
            );
        }
        Matrix3x3<T> &operator-=(const Matrix3x3<T> &other) {
            m_rows[0] -= other.m_rows[0];
            m_rows[1] -= other.m_rows[1];
            m_rows[2] -= other.m_rows[2];
            return *this;
        }

        Matrix3x3<T> &operator*=(const Matrix3x3<T> &other) {
            return (*this = *this * other);
        }
        Matrix3x3<T> operator*(const T &s) const {
            return Matrix3x3(
                m_rows[0] * s,
                m_rows[1] * s,
                m_rows[2] * s
            );
        }
        Matrix3x3<T> &operator*=(const T &s) {
            return (*this = *this * s);
        }

        /*
        Matrix3x3<T> operator/(const Matrix3x3<T> &other) const { return Matrix3x3(m_x / other.m_x, m_y / other.m_y, m_z / other.m_z); }
        Matrix3x3<T> &operator/=(const Matrix3x3<T> &other) { m_x /= other.m_x; m_y /= other.m_y; m_z /= other.m_z; return *this; }
        Matrix3x3<T> operator/(const T &s) const { return Matrix3x3(m_x / s, m_y / s, m_z / s); }
        Matrix3x3<T> &operator/=(const T &s) { m_x /= s; m_y /= s; m_z /= s; return *this; }
        */

        Vector3<T> &operator[](const int idx) { return m_rows[idx]; }
        const Vector3<T> &operator[](const int idx) const { return m_rows[idx]; }

        T &operator()(const int x, const int y) { return m_rows[y][x]; }
        const T &operator()(const int x, const int y) const { return m_rows[y][x]; }

        bool operator==(const Matrix3x3<T> &other) const {
            return  m_rows[0] == other.m_rows[0] &&
                    m_rows[1] == other.m_rows[1] &&
                    m_rows[2] == other.m_rows[2];
        }
        bool operator!=(const Matrix3x3<T> &other) const {
            return  m_rows[0] == other.m_rows[0] ||
                    m_rows[1] != other.m_rows[1] ||
                    m_rows[2] != other.m_rows[2];
        }

        friend std::ostream &operator<<(std::ostream &os, const Matrix3x3<T> &m) {
            os << "Matrix3x3(\n\t" << m[0] << "\n\t" << m[1] << "\n\t" << m[2] << "\n)";
            return os;
        }

    private:
        Vector3<T> m_rows[3];
};

template <typename T>
Matrix3x3<T> operator*(const Matrix3x3<T> &m1, const Matrix3x3<T> &m2) {
    Matrix3x3<T> result(
        {
            m1(0, 0) * m2(0, 0) + m1(1, 0) * m2(0, 1) + m1(2, 0) * m2(0, 2),
            m1(0, 0) * m2(1, 0) + m1(1, 0) * m2(1, 1) + m1(2, 0) * m2(1, 2),
            m1(0, 0) * m2(2, 0) + m1(1, 0) * m2(2, 1) + m1(2, 0) * m2(2, 2)

        },
        {
            m1(0, 1) * m2(0, 0) + m1(1, 1) * m2(0, 1) + m1(2, 1) * m2(0, 2),
            m1(0, 1) * m2(1, 0) + m1(1, 1) * m2(1, 1) + m1(2, 1) * m2(1, 2),
            m1(0, 1) * m2(2, 0) + m1(1, 1) * m2(2, 1) + m1(2, 1) * m2(2, 2)
        },
        {
            m1(0, 2) * m2(0, 0) + m1(1, 2) * m2(0, 1) + m1(2, 2) * m2(0, 2),
            m1(0, 2) * m2(1, 0) + m1(1, 2) * m2(1, 1) + m1(2, 2) * m2(1, 2),
            m1(0, 2) * m2(2, 0) + m1(1, 2) * m2(2, 1) + m1(2, 2) * m2(2, 2)
        }
    );
    return result;
}

template <typename T>
Vector3<T> operator*(const Matrix3x3<T> &m, const Vector3<T> &v) {
    Vector3<T> result({
        m(0, 0) * v[0] + m(1, 0) * v[1] + m(2, 0) * v[2],
        m(0, 1) * v[0] + m(1, 1) * v[1] + m(2, 1) * v[2],
        m(0, 2) * v[0] + m(1, 2) * v[1] + m(2, 2) * v[2]
    });

    return result;
}

template <typename T>
Vector3<T> operator*(const Vector3<T> &v, const Matrix3x3<T> &m) {
    Vector3<T> result({
        m(0, 0) * v[0] + m(0, 1) * v[1] + m(0, 2) * v[2],
        m(1, 0) * v[0] + m(1, 1) * v[1] + m(1, 2) * v[2],
        m(2, 0) * v[0] + m(2, 1) * v[1] + m(2, 2) * v[2]
    });

    return result;
}

